# model.py
import json
import numpy as np
import triton_python_backend_utils as pb_utils
import cv2
import random
import math
from enum import Enum
from torch.utils.dlpack import from_dlpack
import time
import uuid
import os
import torch
import threading
import base64


with open("/home/lcfc/bird_kind_cls/images/imagenet_labels/imagenet_classes_6.txt") as f:
    classes0 = [line.strip() for line in f.readlines()]

with open("/home/lcfc/bird_kind_cls/images/imagenet_labels/names_cls6.txt") as f:
    names0 = eval(f.readlines()[0].strip())

with open("/home/lcfc/bird_kind_cls/images/imagenet_labels/imagenet_classes_1328.txt") as f:
    classes1 = [line.strip() for line in f.readlines()]

with open("/home/lcfc/bird_kind_cls/images/imagenet_labels/names_cls1328.txt") as f:
    names1 = eval(f.readlines()[0].strip())

configs_dic = {}
with open("/home/lcfc/bird_kind_cls/images/configs/model_configs.txt") as f:
    data_list = f.readlines()
    for data in data_list:
        if len(data) == 0:
            continue
        data = data.strip()
        key = data.split(":")[0].strip()
        value = data.split(":")[1].strip()
        if value[0].isnumeric():   # configs
            configs_dic[key] = float(value)
        else:
            configs_dic[key] = value

class inferThread(threading.Thread):
    def __init__(self,triton_self,text):
        threading.Thread.__init__(self)
        self.triton_self = triton_self
        self.text = text

    def run(self):
        self.detect_cls_str = self.triton_self.infer(self.text)

    def get_result(self):
        threading.Thread.join(self)
        try:
            return self.detect_cls_str
        except Exception:
            return json.dumps([])

class TritonPythonModel:
    def initialize(self, args):
        self.model_config = model_config = json.loads(args['model_config'])
        output0_config = pb_utils.get_output_config_by_name(model_config, "output0")
        self.output0_dtype = pb_utils.triton_string_to_numpy(output0_config['data_type'])

    def execute(self, requests):
        output0_dtype = self.output0_dtype
        responses = []
        for request in requests:
            in_0 = pb_utils.get_input_tensor_by_name(request, 'input0')
            in_0 = in_0.as_numpy()   #获取输入字符串的numpy格式数据
            text = in_0[0].decode("utf-8")   #输入的第0个字符，即传入的数据，该数据是客户端编码的格式，需要解码。 '/images/input_img/4.jpg'
            thread1 = inferThread(self,text)
            thread1.start()
            thread1.join()
            detect_cls_str = thread1.get_result()
            out_str_num_name = np.array(detect_cls_str)
            out_tensor_0 = pb_utils.Tensor('output0', out_str_num_name.astype(output0_dtype))
            inference_response = pb_utils.InferenceResponse(output_tensors=[out_tensor_0])
            responses.append(inference_response)
        return responses

    def finalize(self):
        print('Cleaning up...')

    def infer(self,text):
        # read image
        if text[:4] == "/9j/":
            # type 1 is base64
            # base64_str to base64 to opencv numpy
            base64_data = text.encode('utf8')
            img_base64 = base64.b64decode(base64_data)
            np_data = np.frombuffer(img_base64,np.uint8)
            img_arr = cv2.imdecode(np_data,cv2.IMREAD_COLOR)
            draw = img_arr
        else:
            # type 2 is opencv 
            draw = cv2.imread(text)

        if draw is None or draw.shape[-1] == 4:
            return json.dumps([])
         
        if draw.shape[0] < configs_dic["limite_size"] and draw.shape[1] < configs_dic["limite_size"]:
            return json.dumps([])

        # resize image
        draw = self._resize_cls(im0=draw)
        # cls0
        if bool(configs_dic["cls_bool0"]):
            cls_bool1,draw = self._recognize_cls0(draw)
            if cls_bool1 == False:
                return json.dumps([])
        # cls1
        # if bool(configs_dic["cls_bool1"]):
        #     pass
        detect_cls_list = []
        classes_ls1_, predicted_conf_ls1_, category_id_ls1_ = self._recognize_cls1(draw)
        if predicted_conf_ls1_[0] >= configs_dic['cls_conf1']:  # 限制cls的置信度
            tmp_dic = {}
            tmp_dic['object'] = 1
            tmp_dic['box'] = (0,0,0,0)
            tmp_dic['classID'] = category_id_ls1_[0]
            tmp_dic['bird_name'] = classes_ls1_
            tmp_dic['bird_conf'] = predicted_conf_ls1_
            detect_cls_list.append(tmp_dic)
        # save json step 1
        detect_cls_json = json.dumps(detect_cls_list)

        # plot images module
        if bool(configs_dic["plot_bool"]):
            if len(detect_cls_list) != 0:
                # plot text
                cv2.putText(draw, str(classes_ls1_[0]), (10, 50), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 255, 0), 1, cv2.LINE_AA)

            # 文件夹管理2（根据日期建立文件夹，根据时间输出图片和json文件）
            time_now = time.strftime("%Y%m%d", time.localtime())
            output_image_path = '/home/lcfc/bird_kind_cls/images/output_img/' + time_now
            if not os.path.exists(output_image_path):
                os.makedirs(output_image_path)
            # save json step 2
            # output_json_path = '/jsons/' + time_now
            # if not os.path.exists(output_json_path):
            #     os.makedirs(output_json_path)

            uuidFour = str(uuid.uuid4())
            out_pic_name = output_image_path + '/' + uuidFour + '.jpg'
            if draw is not None:
                cv2.imwrite(out_pic_name, draw.astype(np.uint8))
            # save json step 3
            # out_json_name = output_json_path + '/' + uuidFour + '.json'
            # out_pic_json_name = out_pic_name + " " + out_json_name

            #  save json step 4
            # with open(out_json_name,"w") as f:
            #     f.write(detect_cls_json)
        # return out_pic_json_name,detect_cls_json
        return detect_cls_json

    def _recognize_cls0(self,draw):
        input_image_buffer0 = self.preprocess_cls(draw)
        inference_request0 = pb_utils.InferenceRequest(
            model_name=configs_dic['cls_model0'],
            requested_output_names=['prob'],
            inputs=[pb_utils.Tensor('data', input_image_buffer0)]
        )
        inference_response0 = inference_request0.exec()
        result0 = self.pb_tensor_to_numpy(pb_utils.get_output_tensor_by_name(inference_response0, 'prob'))
        result0 = result0.reshape(int(configs_dic['cls_num0']))  # classes num
        classes_ls0, predicted_conf_ls0, category_id_ls0 = self.postprocess_cls(result0,top_k=1,names_mark=names0,classes_mark=classes0)  # [,,,,,],[,,,,,],[,,,,,]
        if category_id_ls0[0] == 0 and predicted_conf_ls0[0] > int(configs_dic['cls_conf0']):
            return True,draw
        else:
            return False,draw

    def _recognize_cls1(self, resize_cls_image):
        input_image_buffer1 = self.preprocess_cls(resize_cls_image)
        inference_request1 = pb_utils.InferenceRequest(
            model_name=configs_dic['cls_model1'],
            requested_output_names=['prob'],
            inputs=[pb_utils.Tensor('data', input_image_buffer1)]
        )
        inference_response1 = inference_request1.exec()
        result1 = self.pb_tensor_to_numpy(pb_utils.get_output_tensor_by_name(inference_response1, 'prob'))
        result1 = result1.reshape(int(configs_dic['cls_num1']))  # classes num
        classes_ls1, predicted_conf_ls1, category_id_ls1 = self.postprocess_cls(result1,top_k=5,names_mark=names1,classes_mark=classes1)  # [,,,,,],[,,,,,],[,,,,,]
        return classes_ls1, predicted_conf_ls1, category_id_ls1

    def _resize_cls(self,im0, new_shape=(224, 224), color=(114, 114, 114)):
        # Resize and pad image while meeting stride-multiple constraints
        shape = im0.shape[:2]  # current shape [height, width]
        # Scale ratio (new / old)
        r = min(new_shape[0] / shape[0], new_shape[1] / shape[1])
        # Compute padding
        new_unpad = int(round(shape[1] * r)), int(round(shape[0] * r))
        dw, dh = new_shape[1] - new_unpad[0], new_shape[0] - new_unpad[1]  # wh padding
        dw /= 2  # divide padding into 2 sides
        dh /= 2
        if shape[::-1] != new_unpad:  # resize
            im0 = cv2.resize(im0, new_unpad, interpolation=cv2.INTER_LINEAR)
        top, bottom = int(round(dh - 0.1)), int(round(dh + 0.1))
        left, right = int(round(dw - 0.1)), int(round(dw + 0.1))
        im0 = cv2.copyMakeBorder(im0, top, bottom, left, right, cv2.BORDER_CONSTANT, value=color)  # add border
        return im0

    def pb_tensor_to_numpy(self,pb_tensor):
        '''pb_tensor转换为numpy格式'''
        if pb_tensor.is_cpu():
            return pb_tensor.as_numpy()
        else:
            pytorch_tensor = from_dlpack(pb_tensor.to_dlpack())
            return pytorch_tensor.cpu().numpy()

    def preprocess_cls(self,input_img):
        mean = (0.485, 0.456, 0.406)
        std = (0.229, 0.224, 0.225)
        im = cv2.cvtColor(input_img, cv2.COLOR_BGR2RGB)
        im = np.float32(im)
        im /= 255.0
        im -= mean
        im /= std
        im = im.transpose(2, 0, 1)
        # prepare batch
        batch_data = np.expand_dims(im, axis=0)
        batch_data = np.ascontiguousarray(batch_data)
        return batch_data

    def postprocess_cls(self,output_data,top_k,names_mark,classes_mark):
        classes_ls = []
        predicted_conf_ls = []
        category_id_ls = []
        output_data = output_data.reshape(1, -1)
        output_data = torch.Tensor(output_data)
        p = torch.nn.functional.softmax(output_data, dim=1)
        score, index = torch.topk(p, top_k)
        # myself add
        remake_list = []
        for re_index in index[0]:
            remake_list.append(int(names_mark[int(re_index)]))
        index = torch.unsqueeze(torch.tensor(remake_list), dim=0)
        for ind in range(index.shape[0]):
            for top_k_i in range(top_k):
                input_category_id = index[ind][top_k_i].item()  # 716
                category_id_ls.append(input_category_id)
                predicted_confidence = score[ind][top_k_i].item()
                predicted_conf_ls.append(predicted_confidence)
                classes_ls.append(classes_mark[input_category_id])
        return classes_ls, predicted_conf_ls, category_id_ls

